/*
 * Copyright (c) 2011 Rozhuk Ivan <rozhuk.im@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */



#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <sys/cpuset.h>
#include <sys/stat.h>
#include <errno.h>
//#include <iostream>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>

#include <sys/stat.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fts.h>
#include <inttypes.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

//#include <default.h>
#define PACKAGE "direct print"
#define VERSION "1.0"







struct prog_settings {
	int			iSktListen;
	unsigned short	usBindPort;
	char			*pPidFileName;
	char			*pPrinterDevName;
	uid_t			pwUID;// user uid
	gid_t			pwGID;// user gid
};


static struct prog_settings *gppsProgSettings;


static void		Usage			(void);
static int		GetOptions		(int argc, char **argv, struct prog_settings *ppsProgSettings);
static void		SigInstall		(void);
static void		SigHandler		(int iSigNum);
static void		MakeDaemon		(void);
static int		WritePID		(char *pPIDFileName);
static int		SetUserAndGroup	(uid_t pwUID, gid_t pwGID);
static int		open_dev		(char *pdevname, int flags);
static void		ProcessPacketLoop	(struct prog_settings *ppsProgSettings);




int
main(int argc, char *argv[]) {
	int iRet;
	struct sockaddr_in siAddr;
	struct prog_settings psProgSettings;

	bzero(&psProgSettings, sizeof(psProgSettings));

	if (GetOptions(argc, argv, &psProgSettings) != 0) {
		Usage();
		return(0);
	}


	psProgSettings.iSktListen = socket(PF_INET, SOCK_STREAM, IPPROTO_IP);
	if (psProgSettings.iSktListen == -1) { // socket created
		fprintf(stderr, "%s: socket() cant create socket: %i %s\n", PACKAGE, errno, strerror(errno));
		return(errno);
	}

	iRet = 1;
	if (setsockopt(psProgSettings.iSktListen, SOL_SOCKET, SO_REUSEADDR, &iRet, sizeof(iRet)) == -1) {
		fprintf(stderr, "%s: setsockopt(): %i %s\n", PACKAGE, errno, strerror(errno));
		return(errno);
	}

	bzero(&siAddr, sizeof(siAddr));
	siAddr.sin_family = PF_INET;
	siAddr.sin_port = htons(psProgSettings.usBindPort);
	siAddr.sin_addr.s_addr = INADDR_ANY;

	iRet = bind(psProgSettings.iSktListen, (struct sockaddr*)&siAddr, sizeof(siAddr));
	if (iRet == -1) {// socket bind to specified port error
		fprintf(stderr, "%s: cant bind() divert socket: %i %s\n", PACKAGE, errno, strerror(errno));
		return(errno);
	}

	if (listen(psProgSettings.iSktListen, 8) == -1) {
		fprintf(stderr, "%s: cant listen(): %i %s\n", PACKAGE, errno, strerror(errno));
		return(errno);
	}

	SigInstall();

	WritePID(psProgSettings.pPidFileName);// Store pid to file
	SetUserAndGroup(psProgSettings.pwUID,psProgSettings.pwGID);// drop rights

	fprintf(stdout, "%s %s: started\n", PACKAGE, VERSION);
	fprintf(stdout, "%s: bind port: %i \n", PACKAGE, psProgSettings.usBindPort);

	// receive and process packets
	gppsProgSettings = &psProgSettings;
	ProcessPacketLoop(&psProgSettings);


	if (psProgSettings.pPidFileName)
		unlink(psProgSettings.pPidFileName); // Remove pid file
	close(psProgSettings.iSktListen);
	fprintf(stdout,"%s: exiting\n", argv[0]);

return(iRet);
}



static void
Usage() {
	fprintf(stderr, "%s %s -- (c) Rozhuk Ivan <rozhuk.im@gmail.com>\n", PACKAGE, VERSION);
	fprintf(stderr, "   BSD licence.  Website: http://rozhuk.org.ru\n");
	fprintf(stderr,
	       "usage: %s [-d] [-b port] [-c CPUnum] [-p port]\n"
		"       [-p PID file] [-u uid|usr -g gid|grp]\n"
	       " -h           usage (this screen)\n"
	       " -d           become daemon\n"
	       " -b port      bind to port\n",PACKAGE);
	fprintf(stderr,
	       " -p PID file  file name to store PID\n"
	       " -u uid|user  change uid\n"
	       " -g gid|group change gid\n");
}



static int
GetOptions(int argc, char **argv, struct prog_settings *ppsProgSettings) {
	unsigned char *pCur;
	
	if (argc < 2)
		return(1);

	for (int i = 1; i < argc; i++) {
		pCur = (unsigned char*)argv[i];

		if (*pCur++ != '-') {
			fprintf(stderr, "invalid option: \"%s\"\n", argv[i]);
			return(1);
		}

		while(*pCur) {
			switch(*pCur++){
			case '?':case 'h':case 'H':
				return(1);
			case 'd':case 'D':
				MakeDaemon();
				break;
			case 'b':case 'B':
				if (*pCur) {
					ppsProgSettings->usBindPort = atol((char*)pCur);
				}else
				if (argv[++i]) {
					ppsProgSettings->usBindPort = atol(argv[i]);
				}else{
					fprintf(stderr, "option \"-b\" requires bind port number\n");
					return(1);
				}
				break;
			case 'p':case 'P':
				if (*pCur) {
					ppsProgSettings->pPidFileName = (char*)pCur;
				}else
				if (argv[++i]) {
					ppsProgSettings->pPidFileName = argv[i];
				}else{
					fprintf(stderr,"option \"-p\" requires pid file name\n");
					return(1);
				}
				break;
			case 't':case 'T':
				if (*pCur) {
					ppsProgSettings->pPrinterDevName = (char*)pCur;
				}else
				if (argv[++i]) {
					ppsProgSettings->pPrinterDevName = argv[i];
				}else{
					fprintf(stderr,"option \"-p\" requires pid file name\n");
					return(1);
				}
				break;
			case 'u':case 'U': {
				char *pUID;

				if (*pCur) {
					pUID = (char*)pCur;
				}else
				if (argv[++i]) {
					pUID = argv[i];
				}else{
					fprintf(stderr, "option \"-u\" requires UID\n");
					return(1);
				}

				if (pUID) {
					struct passwd *pwd;

					pwd = getpwnam(pUID);
					if (pwd) {
						ppsProgSettings->pwUID = pwd->pw_uid;
					}else{
						fprintf(stderr, "option \"-u\" requires UID, UID %s not found\n", pUID);
					}
				}
				}break;
			case 'g':case 'G': {
				char *pGID;

				if (*pCur) {
					pGID = (char*)pCur;
				}else
				if (argv[++i]) {
					pGID = argv[i];
				}else{
					fprintf(stderr, "option \"-g\" requires GID\n");
					return(1);
				}

				if (pGID) {
					struct group *grp;

					grp = getgrnam(pGID);
					if (grp) {
						ppsProgSettings->pwGID = grp->gr_gid;
					}else{
						fprintf(stderr, "option \"-g\" requires GID, GID %s not found\n", pGID);
					}
				}
				}break;
			default:
				fprintf(stderr,"invalid option: \"%c\"\n",*(pCur-1));
				return(1);
			}
		}
	}

return(0);
}



static void
SigInstall() {
	signal(SIGINT,	SigHandler);
	signal(SIGTERM,	SigHandler);
	signal(SIGKILL,	SigHandler);
	signal(SIGHUP,	SigHandler);
	signal(SIGUSR1,	SigHandler);
	signal(SIGUSR2,	SigHandler);
	signal(SIGPIPE,	SIG_IGN);
}



static void
SigHandler(int iSigNum) {

	switch(iSigNum){
	case SIGINT:
	case SIGTERM:
	case SIGKILL:
		close(gppsProgSettings->iSktListen);
		break;
	case SIGHUP:
	case SIGUSR1:
	case SIGUSR2:
		break;
	}
}



static void
MakeDaemon() {

	switch (fork()) {
	case -1:
		fprintf(stderr,"MakeDaemon: fork() failed: %i %s\n", errno, strerror(errno));
		exit(1);
		return;
	case 0: /* child */
		break;
	default: /* parent */
		exit(0);
	}

	// Child
	setsid();
	setpgid(getpid(),0);
	chdir("/tmp/");
	umask(077);

	// Close stdin/stdout/stderr
	close(0);
	close(1);
	close(2);
}



static int
WritePID(char *pPIDFileName) {
	int iFileHandle;

	if (pPIDFileName == NULL)
		return(0);

	iFileHandle = open(pPIDFileName, (O_RDWR|O_CREAT|O_TRUNC));
	if (iFileHandle == -1) {
		fprintf(stderr, "WritePID: fopen() failed: %i %s\n", errno, strerror(errno));
	}else{
		char szPid[64];
		size_t dwPidSize;
			
		dwPidSize = snprintf(szPid, sizeof(szPid), "%d", getpid());
		write(iFileHandle, szPid, dwPidSize);
		fchmod(iFileHandle, (S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH));
		close(iFileHandle);
		return(1);
	}

return(0);
}



static int
SetUserAndGroup(uid_t pwUID,gid_t pwGID) {
	struct passwd *pwd;

	if (pwUID == 0 || pwGID == 0)
		return(0);

	pwd = getpwuid(pwUID);
	if (pwd == NULL) {
		fprintf(stderr, "getpwuid() failed: %i %s\n", errno, strerror(errno));
		return(0);
	}

	if (setgid(pwGID) != 0)
		fprintf(stderr, "setgid() failed: %i %s\n", errno, strerror(errno));
#ifdef HAVE_INITGROUPS
	if (initgroups(pwd->pw_name,pwGID) !=0 )
		fprintf(stderr, "initgroups() failed: %i %s\n", errno, strerror(errno));
#else
#if HAVE_SETGROUPS
	if (setgroups(1, &pwd->pw_gid) != 0)
		fprintf(stderr, "initgroups() failed: %i %s\n", errno, strerror(errno));
#endif
#endif
	if (setuid(pwUID) != 0)
		fprintf(stderr, "setuid() failed: %i %s\n", errno, strerror(errno));

return(1);
}



static int
open_dev(char *pdevname, int flags) {
	int idev;

	idev = open(pdevname, flags);
	if (idev == -1) {
		fprintf(stderr, "open_dev failed: %i %s\n", errno, strerror(errno));
	}
return(idev);
}



static void
UnidirectionalPrint(int iCliSkt, int iPrinterDev) {
	unsigned char buf[(128*1024)];
	ssize_t iReceived, iWrited, iRet;

	while ( (iReceived = recv(iCliSkt, buf, sizeof(buf), 0)) > 0) {
		iWrited = 0;

		fprintf(stdout, "%s: Received %ld\n", PACKAGE, iReceived);

		while (iWrited < iReceived) {
			iRet = write(iPrinterDev, &buf[iWrited], (iReceived - iWrited));
			fprintf(stdout, "%s: Writed %ld\n", PACKAGE, iRet);
			if ( iRet == -1) {
				fprintf(stderr, "%s: cant write: %i %s\n", PACKAGE, errno, strerror(errno));
				return;
			}
			iWrited += iRet;
		}
	}
}



static void
ProcessPacketLoop(struct prog_settings *ppsProgSettings) {
	int iPrinterDev, iCliSkt;
	socklen_t CliAddrLen;
	struct sockaddr_in siCliAddr;

	fprintf(stdout, "%s: Waiting connection...\n", PACKAGE);
	CliAddrLen = sizeof(siCliAddr);
	while ((iCliSkt = accept(ppsProgSettings->iSktListen, (struct sockaddr *)&siCliAddr, &CliAddrLen)) >= 0) {
		fprintf(stdout, "%s: Connection from %s port %hu accepted\n", PACKAGE, inet_ntoa(siCliAddr.sin_addr), ntohs(siCliAddr.sin_port));

		/* Make sure lp device is open... */
		while ((iPrinterDev = open_dev(ppsProgSettings->pPrinterDevName, (O_WRONLY|O_DIRECT))) == -1)
			sleep(10);

		fprintf(stdout, "%s: Printer ready\n", PACKAGE);

		UnidirectionalPrint(iCliSkt, iPrinterDev);

		close(iCliSkt);
		close(iPrinterDev);
		CliAddrLen = sizeof(siCliAddr);

		fprintf(stdout, "%s: Disconnected, waiting next connection...\n", PACKAGE);
	}
}



